// Copyright (c) 2016, Thomas L. Falch
// For conditions of distribution and use, see the accompanying LICENSE and README files

// This file is part of the ImageCL source-to-source compiler
// developed at the Norwegian University of Science and technology


#include "footprintfinder.h"
#include "kernelinfo.h"

using namespace std;


Point::Point(int x, int y, int z) : x(x), y(y), z(z) {}

bool Point::operator<(const Point &lhs) const {
    if (lhs.x == this->x && lhs.y == this->y) {
        return lhs.z < this->z;
    } else if (lhs.x == this->x) {
        return lhs.y < this->y;
    } else {
        return lhs.x < this->x;
    }
}

Offset::Offset() { threadStatic = true; }

Offset::Offset(set<int> values, bool valid) : values(values), valid(valid) { threadStatic = true; }

Offset::Offset(bool valid) : valid(valid) { threadStatic = true; }

Offset::Offset(int i, bool valid) : valid(valid) {
    this->values.insert(i);
    threadStatic = true;
}

Offset Offset::combine(Offset left, Offset right, function<int(int, int)> op) {
    Offset retOffset;
    for (int l : left.values) {
        for (int r : right.values) {
            retOffset.values.insert(op(l, r));
        }
    }
    retOffset.valid = left.valid && right.valid;
    retOffset.threadStatic = left.threadStatic && right.threadStatic;

    return retOffset;
}

string Offset::str() {
    ostringstream s;
    s << "[";
    for (int i : values) {
        s << i << ",";
    }
    if (valid) {
        s << "(valid)]";
    } else {
        s << "(invalid)]";
    }
    return s.str();
}

int HaloSize::getMax() const {
    return max(max(max(up, down), max(left, right)), max(front, back));
}

string HaloSize::str() {
    ostringstream s;
    s << "up: " << up << " down: " << down << " left: " << left << " right: " << right << " front: " << front << " back: " << back;
    return s.str();
}

Footprint::Footprint() {
    threadStatic = true;
    points = new set<Point>();
}

Footprint::Footprint(vector<tuple<Offset, Offset, Offset> > *rawFootprint) {
    threadStatic = true;
    points = new set<Point>();
    for (tuple<Offset, Offset, Offset> rawPoint : *rawFootprint) {
        if (get<0>(rawPoint).valid && get<1>(rawPoint).valid) {
            for (int x : get<0>(rawPoint).values) {
                for (int y : get<1>(rawPoint).values) {
                    for (int z : get<2>(rawPoint).values) {
                        points->insert(Point(x, y, z));
                    }
                }
            }
        }
    }
}

set<Point> *Footprint::getPoints() {
    return this->points;
}

HaloSize Footprint::computeHaloSize() const {
    HaloSize hs;

    for (Point p : *points) {
        if (p.x < 0) {
            if (abs(p.x) > hs.left)
                hs.left = abs(p.x);
        } else {
            if (p.x > hs.right) {
                hs.right = p.x;
            }
        }
        if (p.y < 0) {
            if (abs(p.y) > hs.up)
                hs.up = abs(p.y);
        } else {
            if (p.y > hs.down) {
                hs.down = p.y;
            }
        }
        if (p.z < 0) {
            if (abs(p.z) > hs.front) {
                hs.front = abs(p.z);
            }
        } else {
            if (p.z > hs.back) {
                hs.back = p.z;
            }
        }
    }
    return hs;
}


string Footprint::str() {
    ostringstream s;
    for (Point p : *points) {
        s << "[" << p.x << "," << p.y << "," << p.z << "] ";
    }
    s << "thread-static: " << std::boolalpha << threadStatic << std::noboolalpha;
    return s.str();
}


FootprintAnalysis::FootprintAnalysis(MultiConstantPropagationAnalysis *cpa, KernelInfo *kernelInfo) : cpa(cpa),
                                                                                                      kernelInfo(
                                                                                                              kernelInfo) {}

SgName findArrayName(SgPntrArrRefExp *arrRef) {
    SgNode *child = arrRef->get_lhs_operand();

    while (isSgPntrArrRefExp(child)) {
        arrRef = isSgPntrArrRefExp(child);
        child = arrRef->get_lhs_operand();
    }

    SgVarRefExp *varRef = isSgVarRefExp(child);

    return varRef->get_symbol()->get_name();
}


Offset FootprintAnalysis::fold(SgExpression *expr, FiniteVarsExprsProductLattice *lat) {
    //cout << lat->str() << endl;
    if (const auto *addOp = isSgAddOp(expr)) {
        Offset left = fold(addOp->get_lhs_operand(), lat);
        Offset right = fold(addOp->get_rhs_operand(), lat);

        return Offset::combine(left, right, [](int a, int b) -> int { return a + b; });
    }
    if (const auto *subOp = isSgSubtractOp(expr)) {
        Offset left = fold(subOp->get_lhs_operand(), lat);
        Offset right = fold(subOp->get_rhs_operand(), lat);

        return Offset::combine(left, right, [](int a, int b) -> int { return a - b; });
    }
    if (const auto *mulOp = isSgMultiplyOp(expr)) {
        Offset left = fold(mulOp->get_lhs_operand(), lat);
        Offset right = fold(mulOp->get_rhs_operand(), lat);

        return Offset::combine(left, right, [](int a, int b) -> int { return a * b; });
    }
    if (const auto divOp = isSgDivideOp(expr)) {
        Offset left = fold(divOp->get_lhs_operand(), lat);
        Offset right = fold(divOp->get_rhs_operand(), lat);

        return Offset::combine(left, right, [](int a, int b) -> int { return a / b; });
    }

    if (const auto *intVal = isSgIntVal(expr)) {
        return Offset(intVal->get_value(), true);
    }
    if (auto *varRef = isSgVarRefExp(expr)) {
        SgName varName = varRef->get_symbol()->get_name();
        if (varName == "idx" || varName == "idy" || varName == "idz") {
            Offset o(0, true);
            o.threadStatic = false;
            return o;
        }

        auto *lattice = dynamic_cast<VarsExprsProductLattice *> (NodeState::getLatticeAbove(cpa, varRef, 0)[0]);
        if (!lattice) {
            exit(56);
        }

        set<varID> varIDs = lattice->getAllVars();

        for (const varID &var: varIDs) {
            if (var.str() == varName.str()) {
                auto *constPropLat = dynamic_cast<MultiConstantPropagationLattice *>(lattice->getVarLattice(
                        var));
                if (constPropLat->isConstant()) {
                    return Offset(constPropLat->getValues(), true);
                }
            }
        }
        return Offset(false);
    }
    exit(55);
}


void FootprintAnalysis::visit(const Function &func, const DataflowNode &n, NodeState &state) {
    SgNode *astNode = n.getNode();
    SgNode *parent = astNode->get_parent();

    SgPntrArrRefExp *arrRef = isSgPntrArrRefExp(astNode);

    auto *lat = dynamic_cast<FiniteVarsExprsProductLattice *>(state.getLatticeAbove(cpa)[0]);

    if (arrRef && !isSgPntrArrRefExp(parent)) {
        string name = findArrayName(arrRef);
        const bool is_alias = kernelInfo->isAliasArray(name);
        string aliased;

        if (is_alias) {
            for (const auto &alias : kernelInfo->aliasArrays) {
                if (name == alias.first) {
                    aliased = alias.second.first;
                }
            }
        }

        if (kernelInfo->isImageArray(name) || (is_alias && kernelInfo->isImageArray(aliased))) {
            // All Images are 2d.

            //cout << "[fp] " << arrRef->unparseToString() << ": " << arrRef->get_lhs_operand()->unparseToString()
            //     << ", " << arrRef->get_rhs_operand()->unparseToString() << endl;

            SgPntrArrRefExp *middle = isSgPntrArrRefExp(arrRef->get_lhs_operand());
            SgExpression *x = middle->get_rhs_operand();
            SgExpression *y = arrRef->get_rhs_operand();

            Offset xOffset = fold(x, lat);
            Offset yOffset = fold(y, lat);

            // TODO: Check if this is handled correctly for 2d.
            Offset zOffset(0, true);
            zOffset.threadStatic = true;

            //cout << "[fp] found raw Image footprint: " << name.str() << " : " << xOffset.str() << " ; "
            //     << yOffset.str() << endl;
            auto vec = internalFootprintTable[name];
            if (vec) {
                vec->push_back(make_tuple(xOffset, yOffset, zOffset));
            } else {
                internalFootprintTable[name] = new vector<tuple<Offset, Offset, Offset>>();
                internalFootprintTable[name]->push_back(make_tuple(xOffset, yOffset, zOffset));
            }
        }

        if (kernelInfo->isBlockArray(name) || (is_alias && kernelInfo->isBlockArray(aliased))) {
            // All blocks are 3d.

            //cout << "[fp] " << arrRef->unparseToString() << endl;

            auto *outer_lhs = isSgPntrArrRefExp(arrRef->get_lhs_operand());
            auto *inner_lhs = isSgPntrArrRefExp(outer_lhs->get_lhs_operand());
            auto *x = inner_lhs->get_rhs_operand();
            auto *y = outer_lhs->get_rhs_operand();
            auto *z = arrRef->get_rhs_operand();

            //cout << "[fp] x: " << x->unparseToString() << ", y: " << y->unparseToString() << ", z: " << z->unparseToString() << endl;

            Offset xOffset = fold(x, lat);
            Offset yOffset = fold(y, lat);
            Offset zOffset = fold(z, lat);

            //cout << "[fp] found raw Block footprint: " << name.str() << " : " << xOffset.str() << " ; "
            //     << yOffset.str() << " ; " << zOffset.str() << endl;
            auto vec = internalFootprintTable[name];
            if (vec) {
                vec->push_back(make_tuple(xOffset, yOffset, zOffset));
            } else {
                internalFootprintTable[name] = new vector<tuple<Offset, Offset, Offset>>();
                internalFootprintTable[name]->push_back(make_tuple(xOffset, yOffset, zOffset));
            }
        }
    }
}

map<string, Footprint> FootprintAnalysis::getFootprintTable() {
    map<string, Footprint> footprintTable;

    for (pair<string, vector<tuple<Offset, Offset, Offset>> *> tableEntry : this->internalFootprintTable) {

        bool valid = true;
        bool threadStatic = true;
        for (tuple<Offset, Offset, Offset> p : *tableEntry.second) {
            valid = (valid && get<0>(p).valid && get<1>(p).valid && get<2>(p).valid);
            threadStatic = (threadStatic && get<0>(p).threadStatic && get<1>(p).threadStatic && get<2>(p).threadStatic);
        }

        if (valid) {
            Footprint footprint(tableEntry.second);
            footprint.threadStatic = threadStatic;
            footprintTable[tableEntry.first] = footprint;
        }
    }

    return footprintTable;
}

map<string, Footprint> FootprintFinder::findFootprints(SgProject *project, KernelInfo *kernelInfo) {
    initAnalysis(project);
    // This debug shit is apparently needed to avoid segfaults...
    Dbg::init("test", ".", "index.html");

    LiveDeadVarsAnalysis ldva(project);
    UnstructuredPassInterDataflow upid_ldva(&ldva);
    upid_ldva.runAnalysis();

    // There is a bug or something in the destructor, so we just dont destroy it...
    auto *cgb = new CallGraphBuilder(project);
    cgb->buildCallGraph();
    SgIncidenceDirectedGraph *graph = cgb->getGraph();


    MultiConstantPropagationAnalysis cpa(&ldva);
    ContextInsensitiveInterProceduralDataflow ciipd_cpa(&cpa, graph);
    ciipd_cpa.runAnalysis();

    //Doing this as an analysis turned out to be unneccesary...
    FootprintAnalysis footprintAnalysis(&cpa, kernelInfo);
    UnstructuredPassInterAnalysis upia_footprint(footprintAnalysis);
    upia_footprint.runAnalysis();

    map<string, Footprint> footprintTable = footprintAnalysis.getFootprintTable();

    return footprintTable;
}
