// Copyright (c) 2016, Thomas L. Falch
// For conditions of distribution and use, see the accompanying LICENSE and README files

// This file is part of the ImageCL source-to-source compiler
// developed at the Norwegian University of Science and technology


#include "rose.h"
#include "imagememtransformer.h"
#include "astutil.h"

using namespace std;
using namespace SageBuilder;
using namespace SageInterface;


ImageMemTransformer::ImageMemTransformer(SgProject *project, Parameters params, KernelInfo kernelInfo,
                                         Settings settings) : project(project), params(params), kernelInfo(kernelInfo),
                                                              settings(settings) {
}


void ImageMemTransformer::transform() {
    updateImageMemReferences();
    addSampler();
}


void ImageMemTransformer::updateImageMemReference(SgVarRefExp *varRef, SgScopeStatement *scope) {
    char const *array = varRef->get_symbol()->get_name().str();
    int dimensions = 0;
    SgNode *parent = varRef->get_parent();
    SgNode *temp = varRef;
    while (isSgPntrArrRefExp(parent)) {
        temp = parent;
        parent = parent->get_parent();
        dimensions++;
    }

    SgBinaryOp *top = isSgBinaryOp(temp);
    if (!top) {
        cout << "ERROR: Only indexing access is allowed on texture arrays, found: " << parent->unparseToString() << endl;
        exit(99);
    }
    auto *tmp1 = isSgPntrArrRefExp(top->get_lhs_operand());
    SgExpression *x = nullptr;
    SgExpression *y = nullptr;
    SgExpression *z = nullptr;
    if (kernelInfo.isBlockArray(array)) {
        auto *tmp2 = isSgPntrArrRefExp(tmp1->get_lhs_operand());
        x = tmp2->get_rhs_operand();
        y = tmp1->get_rhs_operand();
        z = top->get_rhs_operand();
    } else {
        x = tmp1->get_rhs_operand();
        y = top->get_rhs_operand();
    }

    // cout << "top: " << top->unparseToString() << endl;
    // cout << "bot: " << tmp1->unparseToString() << endl;
    // cout << "x: " << x->unparseToString() << endl;
    // cout << "y: " << y->unparseToString() << endl;
    // if (z) cout << "z: " << z->unparseToString() << endl;

    const string readImageFunction = [&]() {
        if (kernelInfo.isImageArray(array)) {
            switch (kernelInfo.getPixelType(varRef->get_symbol()->get_name().str())) {
                case FLOAT:
                    return (settings.language == Language::OpenCL) ? "read_imagef" : "tex2D";
                case INT:
                    return (settings.language == Language::OpenCL) ? "read_imagei" : "tex2D";
                case UCHAR:
                    return (settings.language == Language::OpenCL) ? "read_imageui" : "tex2D";
                case DOUBLE:
                    return (settings.language == Language::OpenCL) ? "read_imaged" : "tex2D";
                default:
                    exit(13);
            }
        } else {
            switch (kernelInfo.getPixelType(array)) {
                case FLOAT:
                    return (settings.language == Language::OpenCL) ? "read_imagef" : "tex3D<float>";
                case DOUBLE:
                    /// Double support is not supported for texture memory, so we need to use
                    /// `int2` and cast to double using `__hiloint2double`.
                    /// We generate the `tex3D_double` function later.
                    return (settings.language == Language::OpenCL) ? "read_imaged" : "tex3D_double";
                case INT:
                    return (settings.language == Language::OpenCL) ? "read_imagei" : "tex3D<int>";
                case UCHAR:
                    return (settings.language == Language::OpenCL) ? "read_imageui" : "tex3D<unsigned char>";
                case CHAR:
                    return (settings.language == Language::OpenCL) ? "read_imagei" : "tex3D<char>";
                default:
                    exit(14);
            }
        }
    }();

    auto *readImageCall = [&]() {
       if (settings.language == Language::OpenCL) {
           SgVarRefExp *sampler = buildVarRefExp("sampler", scope);

           SgFunctionCallExp *fakeConversion = nullptr;
           if (kernelInfo.isImageArray(array)) {
               fakeConversion = buildFunctionCallExp("MAKE_INT2", buildIntType(), buildExprListExp(x, y), scope);
           } else {
               fakeConversion = buildFunctionCallExp("MAKE_INT3", buildIntType(), buildExprListExp(x, y, z), scope);
           }

           return buildFunctionCallExp(readImageFunction,
                                buildFloatType(),
                                buildExprListExp(varRef, sampler, fakeConversion),
                                scope);
       } else {
           if (kernelInfo.isImageArray(array)) {
               return buildFunctionCallExp(readImageFunction,
                                           buildFloatType(),
                                           buildExprListExp(buildVarRefExp(string(array) + "_tex", scope), x, y),
                                           scope);
           } else {
               return buildFunctionCallExp(readImageFunction,
                                           buildFloatType(),
                                           buildExprListExp(buildVarRefExp(array, scope), x, y, z),
                                           scope);
           }
       }
    }();

    SgExpression *dotExpression = (settings.language == Language::OpenCL)
                                ? isSgExpression(buildDotExp(readImageCall, buildVarRefExp("x", scope)))
                                : isSgExpression(readImageCall);

    SgBinaryOp *parentBinOp = isSgBinaryOp(parent);
    SgUnaryOp *parentUnaryOp = isSgUnaryOp(parent);
    SgAssignInitializer *parentAssingInit = isSgAssignInitializer(parent);
    if (parentBinOp) {
        if (parentBinOp->get_lhs_operand() == top) {
            parentBinOp->set_lhs_operand(dotExpression);
        } else {
            parentBinOp->set_rhs_operand(dotExpression);
        }
    } else if (parentUnaryOp) {
        parentUnaryOp->set_operand(dotExpression);
    } else if (parentAssingInit) {
        parentAssingInit->set_operand(dotExpression);
    }
}

void ImageMemTransformer::updateImageMemReferences() {
    SgFunctionDeclaration *funcDef = AstUtil::getFunctionDeclaration(project, kernelInfo.getKernelName());

    for (SgNode *varRefNode : NodeQuery::querySubTree(funcDef, V_SgVarRefExp)) {
        for (const string &imgMemArray : *params.imageMemArrays) {
            SgVarRefExp *varRef = isSgVarRefExp(varRefNode);
            if (varRef->get_symbol()->get_name().str() == imgMemArray) {
                updateImageMemReference(varRef, funcDef->get_definition()->get_body());
            }
        }
    }
}

void ImageMemTransformer::addSampler() {
    if (settings.language == Language::OpenCL) {
        auto globals = NodeQuery::querySubTree(project, V_SgGlobal);

        for (SgNode *node : globals) {
            SgGlobal *global = isSgGlobal(node);

            //These are striclty speaking not var refs, but macros?
            auto normCoordsFalse = buildVarRefExp("CLK_NORMALIZED_COORDS_FALSE", global);
            auto addressClampEdge = buildVarRefExp("CLK_ADDRESS_CLAMP", global);
            auto filterNearest = buildVarRefExp("CLK_FILTER_NEAREST", global);

            auto orExpression = buildBitOrOp(normCoordsFalse, buildBitOrOp(filterNearest, addressClampEdge));

            auto initializer = buildAssignInitializer(orExpression, buildIntType());
            SgModifierType *type = buildModifierType(buildOpaqueType("sampler_t", global));
            type->get_typeModifier().setOpenclConstant();
            global->prepend_declaration(buildVariableDeclaration("sampler", type, initializer, global));
        }
    }
}
