// Copyright (c) 2016, Thomas L. Falch
// For conditions of distribution and use, see the accompanying LICENSE and README files

// This file is part of the ImageCL source-to-source compiler
// developed at the Norwegian University of Science and technology


#include "localmemtransformer.h"
#include "astutil.h"

using namespace SageBuilder;
using namespace SageInterface;

LocalMemTransformer::LocalMemTransformer(SgProject *project, Parameters params, KernelInfo kernelInfo,
                                         Settings settings, SgBasicBlock *loopBody) : project(project), params(params),
                                                                                      kernelInfo(kernelInfo),
                                                                                      loopBody(loopBody),
                                                                                      settings(settings) {
}


void LocalMemTransformer::transform() {
    SgFunctionDeclaration *funcDef = AstUtil::getFunctionDeclaration(project, kernelInfo.getKernelName());

    bool first = true;
    for (const auto &localArray : *params.localMemArrays) {
        insertSharedMemLoading(funcDef, localArray, first);
        replaceGlobalWithSharedLoads(localArray);
        prependLocalIdComputation(loopBody, localArray);
        fixVariableReferences(project);
        first = false;
    }
}


SgBasicBlock *
LocalMemTransformer::buildLoadingLoopBody(SgFunctionDeclaration *funcDef, int sm_x, string array, const HaloSize &hs) {
    SgBasicBlock *block = buildBasicBlock();
    auto *scope = funcDef->get_definition()->get_scope();

    const string iterLoader = "smem_load_" + array;

    auto *computeLocalX = buildAssignInitializer(buildModOp(buildVarRefExp(iterLoader, block), buildIntVal(sm_x)));
    auto *computeLocalY = buildAssignInitializer(buildDivideOp(buildVarRefExp(iterLoader, block), buildIntVal(sm_x)));

    auto *localXDecl = buildVariableDeclaration("local_x_" + array, buildConstType(buildIntType()), computeLocalX, block);
    auto *localYDecl = buildVariableDeclaration("local_y_" + array, buildConstType(buildIntType()), computeLocalY, block);

    SgExpression *getGroupIdx = nullptr;
    SgExpression *getGroupIdy = nullptr;
    if (settings.language == Language::OpenCL) {
        getGroupIdx = buildFunctionCallExp("get_group_id", buildIntType(), buildExprListExp(buildIntVal(0)), scope);
        getGroupIdy = buildFunctionCallExp("get_group_id", buildIntType(), buildExprListExp(buildIntVal(1)), scope);
    } else if (settings.language == Language::Cuda) {
        getGroupIdx = buildOpaqueVarRefExp("blockIdx.x", scope);
        getGroupIdy = buildOpaqueVarRefExp("blockIdx.y", scope);
    } else {
        exit(302);
    }

    auto *computeGlobalXTemp = buildMultiplyOp(getGroupIdx, buildIntVal(params.elementsPerThreadX * params.localSizeX));
    auto *computeGlobalYTemp = buildMultiplyOp(getGroupIdy, buildIntVal(params.elementsPerThreadY * params.localSizeY));

    SgExpression *computeGlobalXPadding = buildSubtractOp(buildAddOp(computeGlobalXTemp, buildVarRefExp(localXDecl)), buildIntVal(hs.left));
    SgExpression *computeGlobalYPadding = buildSubtractOp(buildAddOp(computeGlobalYTemp, buildVarRefExp(localYDecl)), buildIntVal(hs.up));

    if ((settings.generateMPI || settings.generateOMP) && kernelInfo.needsMpiBroadcast(array)) {
        computeGlobalXPadding = buildAddOp(computeGlobalXPadding, buildVarRefExp("base_x", scope));
        computeGlobalYPadding = buildAddOp(computeGlobalYPadding, buildVarRefExp("base_y", scope));
    }
    auto *computeGlobalX = buildAssignInitializer(computeGlobalXPadding);
    auto *computeGlobalY = buildAssignInitializer(computeGlobalYPadding);

    auto *globalXDecl = buildVariableDeclaration("global_x_" + array, buildConstType(buildIntType()), computeGlobalX, block);
    auto *globalYDecl = buildVariableDeclaration("global_y_" + array, buildConstType(buildIntType()), computeGlobalY, block);

    block->append_statement(localXDecl);
    block->append_statement(localYDecl);
    block->append_statement(globalXDecl);
    block->append_statement(globalYDecl);

    auto *localBufferX = buildPntrArrRefExp(buildVarRefExp("local_buffer_" + array, block), buildVarRefExp(localXDecl));
    auto *localBufferY = buildPntrArrRefExp(localBufferX, buildVarRefExp(localYDecl));

    auto *globalBufferX = buildPntrArrRefExp(buildVarRefExp(array, block), buildVarRefExp("global_x_" + array, block));
    auto *globalBufferY = buildPntrArrRefExp(globalBufferX, buildVarRefExp("global_y_" + array, block));

    auto *loadAssign = buildExprStatement(buildAssignOp(localBufferY, globalBufferY));

    if (hs.getMax() == 0) {
        string aGridArray = kernelInfo.getAGridArray();
        SgVarRefExp *gX = buildVarRefExp("global_x_" + array, block);
        SgVarRefExp *gY = buildVarRefExp("global_y_" + array, block);

        SgVarRefExp *gsX = buildOpaqueVarRefExp("GS_X", block);
        SgVarRefExp *gsY = buildOpaqueVarRefExp("GS_Y", block);

        SgExpression *isInside = buildOrOp(buildLessThanOp(gX, gsX), buildLessThanOp(gY, gsY));
        SgIfStmt *ifInside = buildIfStmt(isInside, loadAssign, buildNullStatement());
        block->append_statement(ifInside);
    } else {
        block->append_statement(loadAssign);
    }

    return block;
}

SgBasicBlock *LocalMemTransformer::buildLoadingLoopBodyForBlock(SgFunctionDeclaration *funcDef, int sm_x, int sm_y, const string &array, const HaloSize &hs) {
    auto *block = buildBasicBlock();
    auto *scope = funcDef->get_definition()->get_scope();

    const string iterLoader = "smem_load_" + array;

    auto *computeLocalX = buildAssignInitializer(buildModOp(buildModOp(buildVarRefExp(iterLoader, block), buildIntVal(sm_x * sm_y)), buildIntVal(sm_x)));
    auto *computeLocalY = buildAssignInitializer(buildDivideOp(buildModOp(buildVarRefExp(iterLoader, block), buildIntVal(sm_x * sm_y)), buildIntVal(sm_x)));
    auto *computeLocalZ = buildAssignInitializer(buildDivideOp(buildVarRefExp(iterLoader, block), buildIntVal(sm_x * sm_y)));

    auto *localXDecl = buildVariableDeclaration("local_x_" + array, buildConstType(buildIntType()), computeLocalX, block);
    auto *localYDecl = buildVariableDeclaration("local_y_" + array, buildConstType(buildIntType()), computeLocalY, block);
    auto *localZDecl = buildVariableDeclaration("local_z_" + array, buildConstType(buildIntType()), computeLocalZ, block);

    SgExpression *groupIdx = nullptr;
    SgExpression *groupIdy = nullptr;
    SgExpression *groupIdz = nullptr;
    if (settings.language == Language::OpenCL) {
        groupIdx = buildFunctionCallExp("get_group_id", buildIntType(), buildExprListExp(buildIntVal(0)), scope);
        groupIdy = buildFunctionCallExp("get_group_id", buildIntType(), buildExprListExp(buildIntVal(1)), scope);
        groupIdz = buildFunctionCallExp("get_group_id", buildIntType(), buildExprListExp(buildIntVal(2)), scope);
    } else if (settings.language == Language::Cuda) {
        groupIdx = buildOpaqueVarRefExp("blockIdx.x", scope);
        groupIdy = buildOpaqueVarRefExp("blockIdx.y", scope);
        groupIdz = buildOpaqueVarRefExp("blockIdx.z", scope);
    } else {
        exit(301);
    }

    auto *computeGlobalXTmp = buildMultiplyOp(groupIdx, buildIntVal(params.elementsPerThreadX * params.localSizeX));
    auto *computeGlobalYTmp = buildMultiplyOp(groupIdy, buildIntVal(params.elementsPerThreadY * params.localSizeY));
    auto *computeGlobalZTmp = buildMultiplyOp(groupIdz, buildIntVal(params.elementsPerThreadZ * params.localSizeZ));

    SgExpression *computeGlobalXPadding = buildSubtractOp(buildAddOp(computeGlobalXTmp, buildVarRefExp(localXDecl)), buildIntVal(hs.left));
    SgExpression *computeGlobalYPadding = buildSubtractOp(buildAddOp(computeGlobalYTmp, buildVarRefExp(localYDecl)), buildIntVal(hs.up));
    SgExpression *computeGlobalZPadding = buildSubtractOp(buildAddOp(computeGlobalZTmp, buildVarRefExp(localZDecl)), buildIntVal(hs.front));

    if ((settings.generateMPI || settings.generateOMP) && kernelInfo.needsMpiBroadcast(array)) {
        computeGlobalXPadding = buildAddOp(computeGlobalXPadding, buildVarRefExp("base_x", scope));
        computeGlobalYPadding = buildAddOp(computeGlobalYPadding, buildVarRefExp("base_y", scope));
        computeGlobalZPadding = buildAddOp(computeGlobalZPadding, buildVarRefExp("base_z", scope));
    }

    auto *computeGlobalX = buildAssignInitializer(computeGlobalXPadding);
    auto *computeGlobalY = buildAssignInitializer(computeGlobalYPadding);
    auto *computeGlobalZ = buildAssignInitializer(computeGlobalZPadding);

    auto *globalXDecl = buildVariableDeclaration("global_x_" + array, buildConstType(buildIntType()), computeGlobalX, block);
    auto *globalYDecl = buildVariableDeclaration("global_y_" + array, buildConstType(buildIntType()), computeGlobalY, block);
    auto *globalZDecl = buildVariableDeclaration("global_z_" + array, buildConstType(buildIntType()), computeGlobalZ, block);

    block->append_statement(localXDecl);
    block->append_statement(localYDecl);
    block->append_statement(localZDecl);
    block->append_statement(globalXDecl);
    block->append_statement(globalYDecl);
    block->append_statement(globalZDecl);

    auto *localBufferX = buildPntrArrRefExp(buildVarRefExp("local_buffer_" + array, block), buildVarRefExp(localXDecl));
    auto *localBufferY = buildPntrArrRefExp(localBufferX, buildVarRefExp(localYDecl));
    auto *localBufferZ = buildPntrArrRefExp(localBufferY, buildVarRefExp(localZDecl));

    auto *globalBufferX = buildPntrArrRefExp(buildVarRefExp(array, block), buildVarRefExp("global_x_" + array, block));
    auto *globalBufferY = buildPntrArrRefExp(globalBufferX, buildVarRefExp("global_y_" + array, block));
    auto *globalBufferZ = buildPntrArrRefExp(globalBufferY, buildVarRefExp("global_z_" + array, block));

    auto *loadAssign = buildExprStatement(buildAssignOp(localBufferZ, globalBufferZ));

    if (hs.getMax() == 0) {
        auto *gX = buildVarRefExp("global_x_" + array, block);
        auto *gY = buildVarRefExp("global_y_" + array, block);
        auto *gZ = buildVarRefExp("global_z_" + array, block);

        auto *gsX = buildOpaqueVarRefExp("GS_X", block);
        auto *gsY = buildOpaqueVarRefExp("GS_Y", block);
        auto *gsZ = buildOpaqueVarRefExp("GS_Z", block);

        auto *isInside = buildOrOp(buildOrOp(buildLessThanOp(gX, gsX), buildLessThanOp(gY, gsY)), buildLessThanOp(gZ, gsZ));
        auto *ifInside = buildIfStmt(isInside, loadAssign, buildNullStatement());
        block->append_statement(ifInside);
    } else {
        block->append_statement(loadAssign);
    }

    return block;
}

SgFunctionCallExp *LocalMemTransformer::buildBarrierCall(SgScopeStatement *funcScope) {
    if (settings.language == Language::OpenCL) {

        SgExpression *globalMemFence = buildOpaqueVarRefExp("CLK_GLOBAL_MEM_FENCE", getGlobalScope(funcScope));
        SgExpression *localMemFence = buildOpaqueVarRefExp("CLK_LOCAL_MEM_FENCE", getGlobalScope(funcScope));
        SgExpression *memFence = buildBitOrOp(globalMemFence, localMemFence);
        SgFunctionCallExp *barrierCall = buildFunctionCallExp("barrier", buildVoidType(), buildExprListExp(memFence),
                                                              funcScope);
        return barrierCall;
    } else if (settings.language == Language::Cuda) {
        return buildFunctionCallExp("__syncthreads", buildVoidType(), nullptr, funcScope);
    }
    exit(100);
}

SgVariableDeclaration *LocalMemTransformer::buildThreadIdDeclaration(SgScopeStatement *funcScope, string localArray) {
    if (settings.language == Language::OpenCL) {
        SgFunctionCallExp *get_local_id0 = buildFunctionCallExp("get_local_id", buildIntType(),
                                                                buildExprListExp(buildIntVal(0)), funcScope);
        SgFunctionCallExp *get_local_id1 = buildFunctionCallExp("get_local_id", buildIntType(),
                                                                buildExprListExp(buildIntVal(1)), funcScope);
        SgAssignInitializer *threadIdInitializer = buildAssignInitializer(
                buildAddOp(buildMultiplyOp(get_local_id1, buildIntVal(params.localSizeX)), get_local_id0));
        SgVariableDeclaration *threadIdDeclaration = buildVariableDeclaration("thread_id_" + localArray, buildIntType(),
                                                                              threadIdInitializer, funcScope);
        return threadIdDeclaration;
    } else if (settings.language == Language::Cuda) {
        auto *getLocalX = buildOpaqueVarRefExp("threadIdx.x", funcScope);
        auto *getLocalY = buildOpaqueVarRefExp("threadIdx.y", funcScope);
        SgAssignInitializer *threadIdInitializer = buildAssignInitializer(
                buildAddOp(buildMultiplyOp(getLocalY, buildIntVal(params.localSizeX)), getLocalX));
        SgVariableDeclaration *threadIdDeclaration = buildVariableDeclaration("thread_id_" + localArray, buildIntType(),
                                                                              threadIdInitializer, funcScope);
        return threadIdDeclaration;
    }
    exit(101);
}

SgVariableDeclaration *LocalMemTransformer::buildThreadIdDeclarationForBlock(SgScopeStatement *funcScope, const string &array) {
    switch (settings.language) {
        case Language::OpenCL: {
            auto *getLocalX = buildFunctionCallExp("get_local_id", buildIntType(), buildExprListExp(buildIntVal(0)), funcScope);
            auto *getLocalY = buildFunctionCallExp("get_local_id", buildIntType(), buildExprListExp(buildIntVal(1)), funcScope);
            auto *getLocalZ = buildFunctionCallExp("get_local_id", buildIntType(), buildExprListExp(buildIntVal(2)), funcScope);

            auto *threadIdInitializer = buildAssignInitializer(
                    buildAddOp(buildAddOp(buildMultiplyOp(getLocalZ, buildIntVal(params.localSizeY * params.localSizeZ)),
                                          buildMultiplyOp(getLocalY, buildIntVal(params.localSizeX))), getLocalX)
                    );
            return buildVariableDeclaration("thread_id_" + array, buildIntType(), threadIdInitializer, funcScope);
        }
        case Language::Cuda: {
            auto *getLocalX = buildOpaqueVarRefExp("threadIdx.x", funcScope);
            auto *getLocalY = buildOpaqueVarRefExp("threadIdx.y", funcScope);
            auto *getLocalZ = buildOpaqueVarRefExp("threadIdx.z", funcScope);

            auto *threadIdInitializer = buildAssignInitializer(
                    buildAddOp(buildAddOp(buildMultiplyOp(getLocalZ, buildIntVal(params.localSizeY * params.localSizeZ)),
                                          buildMultiplyOp(getLocalY, buildIntVal(params.localSizeX))), getLocalX)
            );
            return buildVariableDeclaration("thread_id_" + array, buildIntType(), threadIdInitializer, funcScope);
        }
        default: exit(102);
    }
}

void LocalMemTransformer::insertSharedMemLoading(SgFunctionDeclaration *funcDef, string localArray, bool withBarrier) {
    Footprint fp;
    try {
        fp = kernelInfo.getFootprintTable().at(localArray);
    } catch (const out_of_range &) {
        // If .at(_) throws, we need to see if `localArray` is aliased.
        for (const auto &alias : kernelInfo.aliasArrays) {
            if (localArray == alias.second.first) {
                fp = kernelInfo.getFootprintTable().at(alias.first);
                break;
            }
        }
    }

    HaloSize hs = fp.computeHaloSize();
    int sharedMemSizeX = params.elementsPerThreadX * params.localSizeX + hs.left + hs.right;
    int sharedMemSizeY = params.elementsPerThreadY * params.localSizeY + hs.up + hs.down;
    int sharedMemSizeZ = params.elementsPerThreadZ * params.localSizeZ + hs.back + hs.front;

    //cout << "[SMEM SIZE] x: " << sharedMemSizeX << ",  y: " << sharedMemSizeY << ", z: " << sharedMemSizeZ << endl;

    SgScopeStatement *funcScope = funcDef->get_definition()->get_scope();

    if (withBarrier) {
        SgFunctionCallExp *barrierCall = buildBarrierCall(funcScope);
        funcDef->get_definition()->get_body()->prepend_statement(buildExprStatement(barrierCall));
    }

    SgType *baseType;
    switch (kernelInfo.getPixelType(localArray)) {
        case FLOAT:
            baseType = buildFloatType();
            break;
        case INT:
            baseType = buildIntType();
            break;
        case DOUBLE:
            baseType = buildDoubleType();
            break;
        default:
            baseType = buildFloatType();
            break;
    }

    if (kernelInfo.isImageArray(localArray)) {
        SgVariableDeclaration *iterDecl = nullptr;
        if (settings.language == Language::OpenCL) {
            iterDecl = buildVariableDeclaration("smem_load_" + localArray, buildIntType(),
                                                buildAssignInitializer(
                                                        buildAddOp(
                                                                buildFunctionCallExp("get_local_id", buildIntType(), buildExprListExp(buildIntVal(0)), funcDef->get_scope()),
                                                                buildMultiplyOp(
                                                                        buildFunctionCallExp("get_local_id", buildIntType(), buildExprListExp(buildIntVal(1)), funcDef->get_scope()),
                                                                        buildFunctionCallExp("get_local_size", buildIntType(), buildExprListExp(buildIntVal(0)), funcDef->get_scope())))),
                                                funcDef->get_scope());
        } else {
            iterDecl = buildVariableDeclaration("smem_load_" + localArray, buildIntType(),
                                                buildAssignInitializer(
                                                        buildAddOp(
                                                                buildOpaqueVarRefExp("threadIdx.x", funcDef->get_scope()),
                                                                buildMultiplyOp(
                                                                        buildOpaqueVarRefExp("threadIdx.y", funcDef->get_scope()),
                                                                        buildOpaqueVarRefExp("blockDim.x", funcDef->get_scope())))),
                                                funcDef->get_scope());
        }

        int sharedMemSize = sharedMemSizeX * sharedMemSizeY;
        auto *condition = buildExprStatement(buildLessThanOp(buildVarRefExp(iterDecl), buildIntVal(sharedMemSize)));
        auto *increment = buildPlusAssignOp(buildVarRefExp(iterDecl), buildIntVal(params.localSizeX * params.localSizeY));

        SgBasicBlock *bodyBlock = buildLoadingLoopBody(funcDef, sharedMemSizeX, localArray, hs);
        bodyBlock->append_statement(buildExprStatement(increment));
        funcDef->get_definition()->get_body()->prepend_statement(buildWhileStmt(condition, bodyBlock));
        funcDef->get_definition()->get_body()->prepend_statement(iterDecl);

        SgType *arrayType = buildArrayType(buildArrayType(baseType, buildIntVal(sharedMemSizeY)), buildIntVal(sharedMemSizeX));
        SgVariableDeclaration *sharedMemDeclaration = nullptr;
        if (settings.language == Language::OpenCL) {
            SgModifierType *modifierType = buildModifierType(arrayType);
            modifierType->get_typeModifier().setOpenclLocal();
            sharedMemDeclaration = buildVariableDeclaration("local_buffer_" + localArray, modifierType, nullptr, funcScope);
        } else if (settings.language == Language::Cuda) {
            sharedMemDeclaration = buildVariableDeclaration("local_buffer_" + localArray, arrayType, nullptr, funcScope);
            sharedMemDeclaration->get_declarationModifier().get_storageModifier().setCudaShared();
        }
        funcDef->get_definition()->get_body()->prepend_statement(sharedMemDeclaration);
    } else if (kernelInfo.isBlockArray(localArray)) {
        SgVariableDeclaration *iterDecl = nullptr;
        if (settings.language == Language::OpenCL) {
            iterDecl = buildVariableDeclaration("smem_load_" + localArray, buildIntType(),
                                                buildAssignInitializer(
                                                        buildAddOp(
                                                                buildFunctionCallExp("get_local_id", buildIntType(), buildExprListExp(buildIntVal(0)), funcDef->get_scope()),
                                                                buildAddOp(
                                                                        buildMultiplyOp(
                                                                                buildFunctionCallExp("get_local_id", buildIntType(), buildExprListExp(buildIntVal(1)), funcDef->get_scope()),
                                                                                buildFunctionCallExp("get_local_size", buildIntType(), buildExprListExp(buildIntVal(0)), funcDef->get_scope())),
                                                                        buildMultiplyOp(
                                                                                buildFunctionCallExp("get_local_id", buildIntType(), buildExprListExp(buildIntVal(2)), funcDef->get_scope()),
                                                                                buildMultiplyOp(
                                                                                        buildFunctionCallExp("get_local_size", buildIntType(), buildExprListExp(buildIntVal(0)), funcDef->get_scope()),
                                                                                        buildFunctionCallExp("get_local_size", buildIntType(), buildExprListExp(buildIntVal(1)), funcDef->get_scope())))))),
                                                funcDef->get_scope());
        } else {
            iterDecl = buildVariableDeclaration("smem_load_" + localArray, buildIntType(),
                                                buildAssignInitializer(
                                                        buildAddOp(
                                                                buildOpaqueVarRefExp("threadIdx.x", funcDef->get_scope()),
                                                                buildAddOp(
                                                                        buildMultiplyOp(
                                                                                buildOpaqueVarRefExp("threadIdx.y", funcDef->get_scope()),
                                                                                buildOpaqueVarRefExp("blockDim.x", funcDef->get_scope())),
                                                                        buildMultiplyOp(
                                                                                buildOpaqueVarRefExp("threadIdx.z", funcDef->get_scope()),
                                                                                buildMultiplyOp(
                                                                                        buildOpaqueVarRefExp("blockDim.x", funcDef->get_scope()),
                                                                                        buildOpaqueVarRefExp("blockDim.y", funcDef->get_scope())))))),
                                                funcDef->get_scope());
        }

        int sharedMemSize = sharedMemSizeX * sharedMemSizeY * sharedMemSizeZ;
        auto *condition = buildExprStatement(buildLessThanOp(buildVarRefExp(iterDecl), buildIntVal(sharedMemSize)));
        auto *increment = buildPlusAssignOp(buildVarRefExp(iterDecl), buildIntVal(params.localSizeX * params.localSizeY * params.localSizeZ));

        auto *bodyBlock = buildLoadingLoopBodyForBlock(funcDef, sharedMemSizeX, sharedMemSizeY, localArray, hs);
        bodyBlock->append_statement(buildExprStatement(increment));
        funcDef->get_definition()->get_body()->prepend_statement(buildWhileStmt(condition, bodyBlock));
        funcDef->get_definition()->get_body()->prepend_statement(iterDecl);

        auto *arrayType = buildArrayType(buildArrayType(buildArrayType(baseType, buildIntVal(sharedMemSizeZ)),
                buildIntVal(sharedMemSizeY)), buildIntVal(sharedMemSizeX));
        SgVariableDeclaration *sharedMemDecl = nullptr;
        switch (settings.language) {
            case Language::OpenCL: {
                auto *modifierType = buildModifierType(arrayType);
                modifierType->get_typeModifier().setOpenclLocal();
                sharedMemDecl = buildVariableDeclaration("local_buffer_" + localArray, modifierType, nullptr, funcScope);
                break;
            }
            case Language::Cuda: {
                sharedMemDecl = buildVariableDeclaration("local_buffer_" + localArray, arrayType, nullptr, funcScope);
                sharedMemDecl->get_declarationModifier().get_storageModifier().setCudaShared();
                break;
            }
            default: exit(55);
        }
        funcDef->get_definition()->get_body()->prepend_statement(sharedMemDecl);
    } else {
        exit(300);
    }
}

// TODO: This is so ugly... find out how to actually do this.
void LocalMemTransformer::replaceWithShared(SgNode *node, string localArray) {
    SgFunctionDeclaration *funcDef = AstUtil::getFunctionDeclaration(project, kernelInfo.getKernelName());
    SgScopeStatement *funcScope = funcDef->get_definition()->get_body();

    cout << "replacing: " << localArray << ": " << node->unparseToString() << endl;

    SgPntrArrRefExp *arrRef = isSgPntrArrRefExp(node);
    if (kernelInfo.isImageArray(localArray)) {
        auto *leftChild = isSgPntrArrRefExp(arrRef->get_lhs_operand());
        leftChild->set_lhs_operand(buildVarRefExp("local_buffer_" + localArray, funcScope));
    } else if (kernelInfo.isBlockArray(localArray)) {
        auto *leftChild = isSgPntrArrRefExp(arrRef->get_lhs_operand());
        auto *leftChild2 = isSgPntrArrRefExp(leftChild->get_lhs_operand());
        leftChild2->set_lhs_operand(buildVarRefExp("local_buffer_" + localArray, funcScope));
    }

    for (SgNode *node2 : NodeQuery::querySubTree(node, V_SgVarRefExp)) {
        SgVarRefExp *varRef = isSgVarRefExp(node2);
        SgNode *parent = varRef->get_parent();
        SgBinaryOp *binOpParent = isSgBinaryOp(parent);

        if (varRef->get_symbol()->get_name() == "idx") {
            if (binOpParent->get_lhs_operand() == varRef) {
                binOpParent->set_lhs_operand(buildVarRefExp("lidx_" + localArray, funcScope));
            }
            if (binOpParent->get_rhs_operand() == varRef) {
                binOpParent->set_rhs_operand(buildVarRefExp("lidx_" + localArray, funcScope));
            }
        }

        if (varRef->get_symbol()->get_name() == "idy") {
            if (binOpParent->get_lhs_operand() == varRef) {
                binOpParent->set_lhs_operand(buildVarRefExp("lidy_" + localArray, funcScope));
            }
            if (binOpParent->get_rhs_operand() == varRef) {
                binOpParent->set_rhs_operand(buildVarRefExp("lidy_" + localArray, funcScope));
            }
        }

        if (varRef->get_symbol()->get_name() == "idz") {
            if (binOpParent->get_lhs_operand() == varRef) {
                binOpParent->set_lhs_operand(buildVarRefExp("lidz_" + localArray, funcScope));
            }
            if (binOpParent->get_rhs_operand() == varRef) {
                binOpParent->set_rhs_operand(buildVarRefExp("lidz_" + localArray, funcScope));
            }
        }
    }
}


bool LocalMemTransformer::isLoadToShared(SgNode *node, const string &localArray) {
    SgNode *parent = node->get_parent();

    SgAssignOp *assignParent = isSgAssignOp(parent);
    if (!assignParent)
        return false;

    SgExpression *lhs = assignParent->get_lhs_operand();

    if (kernelInfo.isImageArray(localArray)) {
        if (!AstUtil::is2DArrayRoot(lhs))
            return false;

        SgPntrArrRefExp *arrRef = isSgPntrArrRefExp(lhs);
        auto *leftChild = dynamic_cast<SgPntrArrRefExp *>(arrRef->get_lhs_operand());
        auto *varRef = dynamic_cast<SgVarRefExp *>(leftChild->get_lhs_operand());

        return varRef->get_symbol()->get_name().str() == ("local_buffer_" + localArray);
    } else if (kernelInfo.isBlockArray(localArray)) {
        if (!AstUtil::is3DArrayRoot(lhs)) {
            return false;
        }
        auto *arrRef = isSgPntrArrRefExp(lhs);
        auto *leftChildX = isSgPntrArrRefExp(arrRef->get_lhs_operand());
        auto *leftChildY = isSgPntrArrRefExp(leftChildX->get_lhs_operand());
        auto *leftChildZ = isSgVarRefExp(leftChildY->get_lhs_operand());
        return leftChildZ->get_symbol()->get_name().str() == ("local_buffer_" + localArray);
    } else {
        return false;
    }
}

void LocalMemTransformer::replaceGlobalWithSharedLoads(string localArray) {
    for (SgNode *node :NodeQuery::querySubTree(project, V_SgPntrArrRefExp)) {
        if (AstUtil::is2DArrayRoot(node) && kernelInfo.isImageArray(localArray)) {
            auto *arrRef = isSgPntrArrRefExp(node);
            auto *leftChild = dynamic_cast<SgPntrArrRefExp *>(arrRef->get_lhs_operand());
            auto *varRef = dynamic_cast<SgVarRefExp *>(leftChild->get_lhs_operand());

            const char * varRefName = varRef->get_symbol()->get_name().str();
            bool isValidAlias = false;
            for (const auto &alias : kernelInfo.aliasArrays) {
                if (varRefName == alias.first) {
                    isValidAlias = localArray == alias.second.first && kernelInfo.isReadOnlyArray(alias.first);
                    if (isValidAlias) break;
                }
            }

            if (varRefName == localArray || isValidAlias) {
                if (isValidAlias || !isLoadToShared(node, localArray)) {
                    replaceWithShared(node, localArray);
                }
            }
        } else if (AstUtil::is3DArrayRoot(node) && kernelInfo.isBlockArray(localArray)) {
            auto *arrRef = isSgPntrArrRefExp(node);
            auto *leftChild = isSgPntrArrRefExp(arrRef->get_lhs_operand());
            auto *leftChild2 = isSgPntrArrRefExp(leftChild->get_lhs_operand());
            auto *varRef = isSgVarRefExp(leftChild2->get_lhs_operand());

            const char * varRefName = varRef->get_symbol()->get_name().str();
            bool isValidAlias = false;
            for (const auto &alias : kernelInfo.aliasArrays) {
                if (varRefName == alias.first) {
                    isValidAlias = localArray == alias.second.first && kernelInfo.isReadOnlyArray(alias.first);
                    if (isValidAlias) break;
                }
            }

            if (varRefName == localArray || isValidAlias) {
                if (isValidAlias || !isLoadToShared(node, localArray)) {
                    replaceWithShared(node, localArray);
                }
            }
        }
    }
}

void LocalMemTransformer::prependLocalIdComputation(SgBasicBlock *loopBody, string localArray) {
    SgScopeStatement *funcScope = nullptr;

    for (SgNode *node : NodeQuery::querySubTree(project, V_SgFunctionDeclaration)) {
        SgFunctionDeclaration *funcDef = isSgFunctionDeclaration(node);
        if (funcDef->get_name().str() == kernelInfo.getKernelName())
            funcScope = funcDef->get_definition()->get_body();
    }

    Footprint fp;
    try {
        fp = kernelInfo.getFootprintTable().at(localArray);
    } catch (const out_of_range &) {
        // If .at(_) throws, we need to see if `localArray` is aliased.
        for (const auto &alias : kernelInfo.aliasArrays) {
            if (localArray == alias.second.first) {
                fp = kernelInfo.getFootprintTable().at(alias.first);
                break;
            }
        }
    }
    if (params.interleaved) {
        if (settings.language == Language::OpenCL) {
            // X
            auto *getLocalIdx = buildFunctionCallExp("get_local_id", buildIntType(),
                                                     buildExprListExp(buildIntVal(0)), funcScope);
            auto *baseLidx = buildAddOp(getLocalIdx, buildMultiplyOp(buildIntVal(params.localSizeX),
                                                                     buildVarRefExp("coars_x", funcScope)));
            auto *lidxWithPadding = buildAddOp(baseLidx, buildIntVal(fp.computeHaloSize().left));
            auto *lidxInit = buildAssignInitializer(lidxWithPadding);
            auto *lidxDecl = buildVariableDeclaration("lidx_" + localArray, buildIntType(),
                                                      lidxInit, funcScope);

            // Y
            auto *getLocalIdy = buildFunctionCallExp("get_local_id", buildIntType(),
                                                     buildExprListExp(buildIntVal(1)), funcScope);
            auto *baseLidy = buildAddOp(getLocalIdy, buildMultiplyOp(buildIntVal(params.localSizeY),
                                                                     buildVarRefExp("coars_y", funcScope)));
            auto *lidyWithPadding = buildAddOp(baseLidy, buildIntVal(fp.computeHaloSize().up));
            auto *lidyInit = buildAssignInitializer(lidyWithPadding);
            auto *lidyDecl = buildVariableDeclaration("lidy_" + localArray, buildIntType(),
                                                      lidyInit, funcScope);

            // Z
            if (kernelInfo.isBlockArray(localArray)) {
                auto *getLocalIdz = buildFunctionCallExp("get_local_id", buildIntType(),
                                                         buildExprListExp(buildIntVal(2)), funcScope);
                auto *baseLidz = buildAddOp(getLocalIdz, buildMultiplyOp(buildIntVal(params.localSizeZ),
                                                                         buildVarRefExp("coars_z", funcScope)));
                auto *lidzWithPadding = buildAddOp(baseLidz, buildIntVal(fp.computeHaloSize().front));
                auto *lidzInit = buildAssignInitializer(lidzWithPadding);
                auto *lidzDecl = buildVariableDeclaration("lidz_" + localArray, buildIntType(),
                                                          lidzInit, funcScope);
                loopBody->prepend_statement(lidzDecl);
            }

            loopBody->prepend_statement(lidyDecl);
            loopBody->prepend_statement(lidxDecl);
        } else if (settings.language == Language::Cuda) {
            // X
            auto *getLocalIdx = buildVarRefExp("threadIdx.x", funcScope);
            auto *baseLidx = buildAddOp(getLocalIdx,
                                        buildMultiplyOp(buildIntVal(params.localSizeX),
                                                        buildVarRefExp("coars_x", funcScope)));
            auto *lidxWithPadding = buildAddOp(baseLidx, buildIntVal(fp.computeHaloSize().left));
            auto *lidxInit = buildAssignInitializer(lidxWithPadding);
            auto *lidxDecl = buildVariableDeclaration("lidx_" + localArray, buildIntType(), lidxInit, funcScope);

            // Y
            auto *getLocalIdy = buildVarRefExp("threadIdx.y", funcScope);
            auto *baseLidy = buildAddOp(getLocalIdy,
                                        buildMultiplyOp(buildIntVal(params.localSizeY),
                                                        buildVarRefExp("coars_y", funcScope)));
            auto *lidyWithPadding = buildAddOp(baseLidy, buildIntVal(fp.computeHaloSize().up));
            auto *lidyInit = buildAssignInitializer(lidyWithPadding);
            auto *lidyDecl = buildVariableDeclaration("lidy_" + localArray, buildIntType(), lidyInit, funcScope);

            // Z
            if (kernelInfo.isBlockArray(localArray)) {
                auto *getLocalIdz = buildVarRefExp("threadIdx.z", funcScope);
                auto *baseLidz = buildAddOp(getLocalIdz,
                                            buildMultiplyOp(buildIntVal(params.localSizeZ),
                                                            buildVarRefExp("coars_z", funcScope)));
                auto *lidzWithPadding = buildAddOp(baseLidz, buildIntVal(fp.computeHaloSize().front));
                auto *lidzInit = buildAssignInitializer(lidzWithPadding);
                auto *lidzDecl = buildVariableDeclaration("lidz_" + localArray, buildIntType(), lidzInit, funcScope);
                loopBody->prepend_statement(lidzDecl);
            }

            loopBody->prepend_statement(lidyDecl);
            loopBody->prepend_statement(lidxDecl);
        }
    } else {
        if (settings.language == Language::OpenCL) {
            // X
            auto *getGlobalIdx = buildFunctionCallExp("get_local_id", buildIntType(),
                                                      buildExprListExp(buildIntVal(0)), funcScope);
            auto *baseLidx = buildAddOp(buildMultiplyOp(getGlobalIdx, buildIntVal(params.elementsPerThreadX)),
                                        buildVarRefExp("coars_x", funcScope));
            auto *lidxWithPadding = buildAddOp(baseLidx, buildIntVal(fp.computeHaloSize().left));
            auto *lidxInit = buildAssignInitializer(lidxWithPadding);
            auto *lidxDecl = buildVariableDeclaration("lidx_" + localArray, buildIntType(),
                                                      lidxInit, funcScope);

            // Y
            auto *getGlobalIdy = buildFunctionCallExp("get_local_id", buildIntType(),
                                                      buildExprListExp(buildIntVal(1)), funcScope);
            auto *baseLidy = buildAddOp(buildMultiplyOp(getGlobalIdy, buildIntVal(params.elementsPerThreadY)),
                                        buildVarRefExp("coars_y", funcScope));
            auto *lidyWithPadding = buildAddOp(baseLidy, buildIntVal(fp.computeHaloSize().up));
            auto *lidyInit = buildAssignInitializer(lidyWithPadding);
            auto *lidyDecl = buildVariableDeclaration("lidy_" + localArray, buildIntType(), lidyInit, funcScope);

            // Z
            if (kernelInfo.isBlockArray(localArray)) {
                auto *getGlobalIdz = buildFunctionCallExp("get_local_id", buildIntType(),
                                                          buildExprListExp(buildIntVal(2)), funcScope);
                auto *baseLidz = buildAddOp(buildMultiplyOp(getGlobalIdz, buildIntVal(params.elementsPerThreadZ)),
                                            buildVarRefExp("coars_z", funcScope));
                auto *lidzWithPadding = buildAddOp(baseLidz, buildIntVal(fp.computeHaloSize().front));
                auto *lidzInit = buildAssignInitializer(lidzWithPadding);
                auto *lidzDecl = buildVariableDeclaration("lidz_" + localArray, buildIntType(), lidzInit, funcScope);
                loopBody->prepend_statement(lidzDecl);
            }

            loopBody->prepend_statement(lidyDecl);
            loopBody->prepend_statement(lidxDecl);
        } else if (settings.language == Language::Cuda) {
            // X
            auto *getGlobalIdx = buildVarRefExp("threadIdx.x", funcScope);
            auto baseLidx = buildAddOp(buildMultiplyOp(getGlobalIdx, buildIntVal(params.elementsPerThreadX)),
                                       buildVarRefExp("coars_x", funcScope));
            auto *lidxWithPadding = buildAddOp(baseLidx, buildIntVal(fp.computeHaloSize().left));
            auto *lidxInit = buildAssignInitializer(lidxWithPadding);
            auto *lidxDecl = buildVariableDeclaration("lidx_" + localArray, buildIntType(), lidxInit, funcScope);

            // Y
            auto *getGlobalIdy = buildVarRefExp("threadIdx.y", funcScope);
            auto baseLidy = buildAddOp(buildMultiplyOp(getGlobalIdy, buildIntVal(params.elementsPerThreadY)),
                                       buildVarRefExp("coars_y", funcScope));
            auto *lidyWithPadding = buildAddOp(baseLidy, buildIntVal(fp.computeHaloSize().up));
            auto *lidyInit = buildAssignInitializer(lidyWithPadding);
            auto *lidyDecl = buildVariableDeclaration("lidy_" + localArray, buildIntType(), lidyInit, funcScope);

            // Z
            if (kernelInfo.isBlockArray(localArray)) {
                auto *getGlobalIdz = buildVarRefExp("threadIdx.z", funcScope);
                auto baseLidz = buildAddOp(buildMultiplyOp(getGlobalIdz, buildIntVal(params.elementsPerThreadZ)),
                                           buildVarRefExp("coars_z", funcScope));
                auto *lidzWithPadding = buildAddOp(baseLidz, buildIntVal(fp.computeHaloSize().front));
                auto *lidzInit = buildAssignInitializer(lidzWithPadding);
                auto *lidzDecl = buildVariableDeclaration("lidz_" + localArray, buildIntType(), lidzInit, funcScope);
                loopBody->prepend_statement(lidzDecl);
            }

            loopBody->prepend_statement(lidyDecl);
            loopBody->prepend_statement(lidxDecl);
        }
    }
}
