#ifndef IMAGECL_REGISTER_CACHING_H
#define IMAGECL_REGISTER_CACHING_H

#include <iostream>
#include <rose.h>
#include "astutil.h"
#include "footprintfinder.h"

using namespace SageBuilder;
using namespace SageInterface;

/// Adds register caching for stencil patterns.
/// Based on `Register Caching for Stencil Compuations on GPUs` by Thomas L. Falch and Anne C. Elster,
/// and https://devblogs.nvidia.com/parallelforall/register-cache-warp-cuda/ from NVIDIA.
class RegisterCaching {
public:
    virtual void transform() = 0;
};

class CudaRegisterCaching : public RegisterCaching {
    SgProject *project = nullptr;
    KernelInfo kernelInfo;
    const Parameters &params;
    SgBasicBlock *loopBody;
    const Settings &settings;

public:
    CudaRegisterCaching(SgProject *project, KernelInfo kernelInfo, const Parameters &params,
                        const Settings &settings, SgBasicBlock *loopBody)
            : RegisterCaching(),
              project(project),
              kernelInfo(std::move(kernelInfo)),
              params(params),
              loopBody(loopBody),
              settings(settings) {}

    inline void transform() override {
        auto *funcDef = AstUtil::getFunctionDeclaration(project, kernelInfo.getKernelName());
        for (const auto &array : params.regCacheArrays) {
            insertRegisterLoading(funcDef, array);
            replaceGlobalLoads(array);
            prependLocalIdComputation(loopBody, array);
            fixVariableReferences(project);
        }
    }

private:
    inline void insertRegisterLoading(SgFunctionDeclaration *funcDef, const std::string &array) {
        Footprint fp;
        try {
            fp = kernelInfo.getFootprintTable().at(array);
        } catch (const out_of_range &) {
            // If .at(_) throws, we need to see if `localArray` is aliased.
            for (const auto &alias : kernelInfo.aliasArrays) {
                if (array == alias.second.first) {
                    fp = kernelInfo.getFootprintTable().at(alias.first);
                    break;
                }
            }
        }

        const HaloSize haloSize = fp.computeHaloSize();
        // Local register cache size.
        const int regCacheSizeX = params.elementsPerThreadX + haloSize.left + haloSize.right; // (haloSize.left > haloSize.right ?  haloSize.left * 2 : haloSize.right * 2);
        const int regCacheSizeY = params.elementsPerThreadY + haloSize.up + haloSize.down; // (haloSize.up > haloSize.down ? haloSize.up * 2 : haloSize.down * 2);
        const int regCacheSizeZ = params.elementsPerThreadZ + haloSize.front + haloSize.back;
        const int regCacheSize = [&]() {
            if (kernelInfo.isImageArray(array)) {
                return regCacheSizeX * regCacheSizeY;
            } else {
                return regCacheSizeX * regCacheSizeY * regCacheSizeZ;
            }
        }();

        auto *funcScope = funcDef->get_definition()->get_scope();

        const std::string iterVar = "reg_cache_load_i_" + array;
        // (X; _; _) { _ }
        auto *init = buildVariableDeclaration(iterVar, buildIntType(), buildAssignInitializer(buildIntVal(0)), funcScope);
        // (_; X; _) { _ }
        auto *test = buildExprStatement(buildLessThanOp(buildVarRefExp(iterVar, funcScope),
                                                        buildIntVal(regCacheSize)));
        // (_; _; X) { _ }
        auto *inc = buildPlusAssignOp(buildVarRefExp(iterVar), buildIntVal(1));

        // (_; _; _) { X }
        auto *body = buildLoadingLoopBody(funcScope, array, iterVar, regCacheSizeX, regCacheSizeY, haloSize);
        body->append_statement(buildExprStatement(inc));
        funcDef->get_definition()->get_body()->prepend_statement(buildWhileStmt(test, body));

        // Build register loader.
        funcDef->get_definition()->get_body()->prepend_statement(init);

        // Lane id.
        SgAssignInitializer *laneInit = nullptr;
        if (kernelInfo.isImageArray(array)) {
            laneInit = buildAssignInitializer(
                    buildModOp(
                            buildAssignInitializer(
                                    buildAddOp(
                                            buildMultiplyOp(
                                                    buildOpaqueVarRefExp("threadIdx.y", funcScope),
                                                    buildOpaqueVarRefExp("blockDim.x", funcScope)
                                            ),
                                            buildOpaqueVarRefExp("threadIdx.x", funcScope)
                                    )
                            ),
                            buildOpaqueVarRefExp("warpSize", funcScope)
                    )
            );
        } else if (kernelInfo.isBlockArray(array)) {
            laneInit = buildAssignInitializer(
                    buildModOp(
                            buildAddOp(
                                    buildAddOp(
                                            buildMultiplyOp(
                                                    buildOpaqueVarRefExp("threadIdx.z", funcScope),
                                                    buildMultiplyOp(buildOpaqueVarRefExp("blockDim.y", funcScope), buildOpaqueVarRefExp("blockDim.x", funcScope))
                                                    ),
                                            buildMultiplyOp(
                                                    buildOpaqueVarRefExp("threadIdx.y", funcScope),
                                                    buildOpaqueVarRefExp("blockDim.x", funcScope)
                                                    )
                                            ),
                                    buildOpaqueVarRefExp("threadIdx.x", funcScope)
                                    ),
                            buildOpaqueVarRefExp("warpSize", funcScope)
                            )
                    );
        }
        auto *lane = buildVariableDeclaration("lane_id_" + array, buildConstType(buildIntType()), laneInit, funcScope);
        funcDef->get_definition()->get_body()->prepend_statement(lane);

        SgType *baseType;
        switch (kernelInfo.getPixelType(array)) {
            case FLOAT:
                baseType = buildFloatType();
                break;
            case INT:
                baseType = buildIntType();
                break;
            case DOUBLE:
                baseType = buildDoubleType();
                break;
            default:
                baseType = buildFloatType();
                break;
        }
        SgType *arrayType = nullptr;
        if (kernelInfo.isImageArray(array)) {
            arrayType = buildArrayType(buildArrayType(baseType, buildIntVal(regCacheSizeY)),
                                       buildIntVal(regCacheSizeX));
        } else if (kernelInfo.isBlockArray(array)) {
            arrayType = buildArrayType(buildArrayType(buildArrayType(baseType, buildIntVal(regCacheSizeY)),
                                                      buildIntVal(regCacheSizeX)), buildIntVal(regCacheSizeZ));
        }
        auto *regCacheDecl = buildVariableDeclaration("reg_cache_" + array, arrayType, nullptr, funcScope);
        funcDef->get_definition()->get_body()->prepend_statement(regCacheDecl);
    }

    inline SgBasicBlock *buildLoadingLoopBody(SgScopeStatement *funcScope, const std::string &array, const std::string &iterVar, int regCacheSizeX, int regCacheSizeY, const HaloSize &hs) {
        auto *block = buildBasicBlock();

        SgVariableDeclaration *x = nullptr, *y = nullptr, *z = nullptr;

        if (kernelInfo.isImageArray(array)) {
            auto *computeX = buildAssignInitializer(buildModOp(buildVarRefExp(iterVar, block), buildIntVal(regCacheSizeX)));
            auto *computeY = buildAssignInitializer(buildDivideOp(buildVarRefExp(iterVar, block), buildIntVal(regCacheSizeX)));

            x = buildVariableDeclaration("x", buildConstType(buildIntType()), computeX, block);
            y = buildVariableDeclaration("y", buildConstType(buildIntType()), computeY, block);
        } else {
            auto *computeZ = buildAssignInitializer(buildDivideOp(buildVarRefExp(iterVar, block), buildIntVal(regCacheSizeX * regCacheSizeY)));
            z = buildVariableDeclaration("z", buildConstType(buildIntType()), computeZ, block);

            auto *iterLoaderTmp = buildAssignInitializer(
                    buildSubtractOp(buildVarRefExp(iterVar, block),
                                    buildMultiplyOp(buildVarRefExp(z), buildIntVal(regCacheSizeX * regCacheSizeY))));
            auto *iterLoaderTmpDecl = buildVariableDeclaration("tmp_reg_load_" + array, buildConstType(buildIntType()), iterLoaderTmp, block);
            block->append_statement(iterLoaderTmpDecl);

            auto *computeX = buildAssignInitializer(buildDivideOp(buildVarRefExp(iterLoaderTmpDecl), buildIntVal(regCacheSizeX)));
            auto *computeY = buildAssignInitializer(buildModOp(buildVarRefExp(iterLoaderTmpDecl), buildIntVal(regCacheSizeX)));

            x = buildVariableDeclaration("x", buildConstType(buildIntType()), computeX, block);
            y = buildVariableDeclaration("y", buildConstType(buildIntType()), computeY, block);
        }

        auto *globalXInit = buildAssignInitializer(
                buildSubtractOp(
                        buildAddOp(
                                buildAddOp(
                                        buildMultiplyOp(
                                            buildOpaqueVarRefExp("blockIdx.x", funcScope),
                                            buildOpaqueVarRefExp("blockDim.x", funcScope)
                                        ),
                                        buildOpaqueVarRefExp("threadIdx.x", funcScope)
                                ),
                                buildVarRefExp(x)
                        ),
                        buildIntVal(hs.left)
                )
        );
        auto *globalXDecl = buildVariableDeclaration("glob_x", buildConstType(buildIntType()), globalXInit, funcScope);
        block->append_statement(x);
        block->append_statement(globalXDecl);

        auto *globalYInit = buildAssignInitializer(
                buildSubtractOp(
                        buildAddOp(
                                buildAddOp(
                                        buildMultiplyOp(
                                                buildOpaqueVarRefExp("blockIdx.y", funcScope),
                                                buildOpaqueVarRefExp("blockDim.y", funcScope)
                                        ),
                                        buildOpaqueVarRefExp("threadIdx.y", funcScope)
                                ),
                                buildVarRefExp(y)
                        ),
                        buildIntVal(hs.up)
                )
        );

        auto *globalYDecl = buildVariableDeclaration("glob_y", buildConstType(buildIntType()), globalYInit, funcScope);
        block->append_statement(y);
        block->append_statement(globalYDecl);

        if (z) {
            auto *globalZInit = buildAssignInitializer(
                    buildSubtractOp(
                            buildAddOp(
                                    buildAddOp(
                                            buildMultiplyOp(
                                                    buildOpaqueVarRefExp("blockIdx.z", funcScope),
                                                    buildOpaqueVarRefExp("blockDim.z", funcScope)
                                            ),
                                            buildOpaqueVarRefExp("threadIdx.z", funcScope)
                                    ),
                                    buildVarRefExp(z)
                            ),
                            buildIntVal(hs.front)
                    )
            );
            auto *globalZDecl = buildVariableDeclaration("glob_z", buildConstType(buildIntType()), globalZInit, funcScope);
            block->append_statement(z);
            block->append_statement(globalZDecl);
        }

        auto *localBufferX = buildPntrArrRefExp(buildVarRefExp("reg_cache_" + array, block), buildVarRefExp("x", block));
        SgPntrArrRefExp *localBuffer = nullptr;
        if (kernelInfo.isImageArray(array)) {
            localBuffer = buildPntrArrRefExp(localBufferX, buildVarRefExp("y", block));
        } else {
            auto *localBufferY = buildPntrArrRefExp(localBufferX, buildVarRefExp("y", block));
            localBuffer = buildPntrArrRefExp(localBufferY, buildVarRefExp("z", block));
        }

        auto *globalBufferX = buildPntrArrRefExp(buildVarRefExp(array, block), buildVarRefExp("glob_x", block));
        SgPntrArrRefExp *globalBuffer = nullptr;
        if (kernelInfo.isImageArray(array)) {
            globalBuffer = buildPntrArrRefExp(globalBufferX, buildVarRefExp("glob_y", block));
        } else {
            auto *globalBufferY = buildPntrArrRefExp(globalBufferX, buildVarRefExp("glob_y", block));
            globalBuffer = buildPntrArrRefExp(globalBufferY, buildVarRefExp("glob_z", block));
        }

        auto *loadAssign = buildExprStatement(buildAssignOp(localBuffer, globalBuffer));

        if (hs.getMax() == 0) {
            const std::string aGridArray = kernelInfo.getAGridArray();
            auto *gX = buildVarRefExp("glob_x", block);
            auto *gY = buildVarRefExp("glob_y", block);
            auto *gZ = buildVarRefExp("glob_z", block);

            auto *gsX = buildOpaqueVarRefExp("GS_X", block);
            auto *gsY = buildOpaqueVarRefExp("GS_Y", block);
            auto *gsZ = buildOpaqueVarRefExp("GS_Z", block);

            auto *isInside = buildOrOp(buildOrOp(buildLessThanOp(gX, gsX), buildLessThanOp(gY, gsY)), buildLessThanOp(gZ, gsZ));
            auto *ifInside = buildIfStmt(isInside, loadAssign, buildNullStatement());
            block->append_statement(ifInside);
        } else {
            block->append_statement(loadAssign);
        }

        return block;
    }

    inline void replaceGlobalLoads(const std::string &array) {
        for (auto *node : NodeQuery::querySubTree(project, V_SgPntrArrRefExp)) {
            if (kernelInfo.isImageArray(array) && AstUtil::is2DArrayRoot(node)) {
                auto *arrRef = isSgPntrArrRefExp(node);
                auto *left = isSgPntrArrRefExp(arrRef->get_lhs_operand());
                auto *varRef = isSgVarRefExp(left->get_lhs_operand());

                const char * varRefName = varRef->get_symbol()->get_name().str();
                bool isValidAlias = false;
                for (const auto &alias : kernelInfo.aliasArrays) {
                    if (varRefName == alias.first) {
                        isValidAlias = array == alias.second.first;
                        break;
                    }
                }

                if (varRefName == array || isValidAlias) {
                    if (isValidAlias || !isLoadToReg(node, array)) {
                        if (kernelInfo.isReadOnlyArray(varRefName)) {
                            replaceWithRegCache(arrRef, array);
                        }
                    }
                }
            } else if (kernelInfo.isBlockArray(array) && AstUtil::is3DArrayRoot(node)) {
                auto *arrRef = isSgPntrArrRefExp(node);
                auto *leftChild = isSgPntrArrRefExp(arrRef->get_lhs_operand());
                auto *leftChild2 = isSgPntrArrRefExp(leftChild->get_lhs_operand());
                auto *varRef = isSgVarRefExp(leftChild2->get_lhs_operand());

                const char * varRefName = varRef->get_symbol()->get_name().str();
                bool isValidAlias = false;
                for (const auto &alias : kernelInfo.aliasArrays) {
                    if (varRefName == alias.first) {
                        isValidAlias = array == alias.second.first;
                        break;
                    }
                }

                if (varRefName == array || isValidAlias) {
                    if (isValidAlias || !isLoadToReg(node, array)) {
                        if (kernelInfo.isReadOnlyArray(varRefName)) {
                            replaceWithRegCache(arrRef, array);
                        }
                    }
                }
            }
        }
    }

    inline bool isLoadToReg(SgNode *node, const std::string &array) {
        if (auto *parent = isSgAssignOp(node->get_parent())) {
            auto *lhs = parent->get_lhs_operand();
            if (kernelInfo.isImageArray(array) && AstUtil::is2DArrayRoot(lhs)) {
                auto *arrRef = isSgPntrArrRefExp(lhs);
                auto *left = isSgPntrArrRefExp(arrRef->get_lhs_operand());
                auto *varRef = isSgVarRefExp(left->get_lhs_operand());
                return varRef->get_symbol()->get_name().str() == ("reg_cache_" + array);
            } else if (kernelInfo.isBlockArray(array) && AstUtil::is3DArrayRoot(lhs)) {
                auto *arrRef = isSgPntrArrRefExp(lhs);
                auto *leftChildX = isSgPntrArrRefExp(arrRef->get_lhs_operand());
                auto *leftChildY = isSgPntrArrRefExp(leftChildX->get_lhs_operand());
                auto *leftChildZ = isSgVarRefExp(leftChildY->get_lhs_operand());
                return leftChildZ->get_symbol()->get_name().str() == ("reg_cache_" + array);
            }
        }
        return false;
    }

    inline void replaceWithRegCache(SgPntrArrRefExp *arrRef, const std::string &array) {
        auto *funcDef = AstUtil::getFunctionDeclaration(project, kernelInfo.getKernelName());
        auto *funcScope = funcDef->get_definition()->get_body();

        if (kernelInfo.isImageArray(array)) {
            auto *leftChild = isSgPntrArrRefExp(arrRef->get_lhs_operand());
            leftChild->set_lhs_operand(buildVarRefExp("reg_cache_" + array, funcScope));
        } else if (kernelInfo.isBlockArray(array)) {
            auto *leftChild = isSgPntrArrRefExp(arrRef->get_lhs_operand());
            auto *leftChild2 = isSgPntrArrRefExp(leftChild->get_lhs_operand());
            leftChild2->set_lhs_operand(buildVarRefExp("reg_cache_" + array, funcScope));
        }

        SgExpression *offsetX = nullptr;
        SgExpression *offsetY = nullptr;
        SgExpression *offsetZ = nullptr;

        auto copy_help = SgTreeCopy();
        auto *arrRef_copy = isSgPntrArrRefExp(arrRef->copy(copy_help));
        const auto fixup_arrRef = [&](SgPntrArrRefExp *arrRef, bool remove_offset) {
            for (auto *node : NodeQuery::querySubTree(arrRef, V_SgVarRefExp)) {
                auto *varRef = isSgVarRefExp(node);
                auto *parent = varRef->get_parent();
                auto *binOpParent = isSgBinaryOp(parent);

                if (varRef->get_symbol()->get_name() == "idx") {
                    if (binOpParent->get_lhs_operand() == varRef) {
                        binOpParent->set_lhs_operand(buildVarRefExp("lidx_" + array, funcScope));
                        if (isSgAddOp(binOpParent) && remove_offset) {
                            offsetX = binOpParent->get_rhs_operand();
                            binOpParent->set_rhs_operand(buildIntVal(0));
                        }
                    }
                    if (binOpParent->get_rhs_operand() == varRef) {
                        binOpParent->set_rhs_operand(buildVarRefExp("lidx_" + array, funcScope));
                        if (isSgAddOp(binOpParent) && remove_offset) {
                            offsetX = binOpParent->get_lhs_operand();
                            binOpParent->set_lhs_operand(buildIntVal(0));
                        }
                    }
                }

                if (varRef->get_symbol()->get_name() == "idy") {
                    if (binOpParent->get_lhs_operand() == varRef) {
                        binOpParent->set_lhs_operand(buildVarRefExp("lidy_" + array, funcScope));
                        if (isSgAddOp(binOpParent) && remove_offset) {
                            offsetY = binOpParent->get_rhs_operand();
                            binOpParent->set_rhs_operand(buildIntVal(0));
                        }
                    }
                    if (binOpParent->get_rhs_operand() == varRef) {
                        binOpParent->set_rhs_operand(buildVarRefExp("lidy_" + array, funcScope));
                        if (isSgAddOp(binOpParent) && remove_offset) {
                            offsetY = binOpParent->get_lhs_operand();
                            binOpParent->set_lhs_operand(buildIntVal(0));
                        }
                    }
                }

                if (varRef->get_symbol()->get_name() == "idz") {
                    if (binOpParent->get_lhs_operand() == varRef) {
                        binOpParent->set_lhs_operand(buildVarRefExp("lidz_" + array, funcScope));
                        if (isSgAddOp(binOpParent) && remove_offset) {
                            offsetZ = binOpParent->get_rhs_operand();
                            binOpParent->set_rhs_operand(buildIntVal(0));
                        }
                    }
                    if (binOpParent->get_rhs_operand() == varRef) {
                        binOpParent->set_rhs_operand(buildVarRefExp("lidz_" + array, funcScope));
                        if (isSgAddOp(binOpParent) && remove_offset) {
                            offsetZ = binOpParent->get_lhs_operand();
                            binOpParent->set_lhs_operand(buildIntVal(0));
                        }
                    }
                }
            }
        };

        fixup_arrRef(arrRef, true);
        fixup_arrRef(arrRef_copy, false);

        if (auto *parent = isSgExpression(arrRef->get_parent())) {
            SgType *baseType, *baseType2;
            switch (kernelInfo.getPixelType(array)) {
                case FLOAT:
                    baseType = buildFloatType();
                    baseType2 = buildFloatType();
                    break;
                case INT:
                    baseType = buildIntType();
                    baseType2 = buildIntType();
                    break;
                case DOUBLE:
                    baseType = buildDoubleType();
                    baseType2 = buildDoubleType();
                    break;
                default:
                    baseType = buildFloatType();
                    baseType2 = buildFloatType();
                    break;
            }

            // (lane_id_input + (i + j * localSizeX + k * localSizeX * localSizeY)) % warpSize
            SgExpression *lane = nullptr;
            if (offsetX && offsetY && offsetZ) {
                lane = buildMultiplyOp(offsetZ, buildIntVal(min(32, params.localSizeX) * min(32, params.localSizeY)));
                lane = buildAddOp(lane, buildMultiplyOp(offsetY, buildIntVal(min(32, params.localSizeX))));
                lane = buildAddOp(lane, offsetX);
            } else if (offsetY && offsetX) {
                lane = buildMultiplyOp(offsetY, buildIntVal(min(32, params.localSizeX)));
                lane = buildAddOp(lane, offsetX);
            } else if (offsetX && offsetZ) {
                lane = buildMultiplyOp(offsetZ, buildIntVal(min(32, params.localSizeX) * min(32, params.localSizeY)));
                lane = buildAddOp(lane, offsetX);
            } else if (offsetY && offsetZ) {
                lane = buildMultiplyOp(offsetZ, buildIntVal(min(32, params.localSizeX) * min(32, params.localSizeY)));
                lane = buildAddOp(lane, buildMultiplyOp(offsetY, buildIntVal(min(32, params.localSizeX))));
            } else if (offsetX) {
                lane = offsetX;
            } else if (offsetY) {
                lane = buildMultiplyOp(offsetY, buildIntVal(min(32, params.localSizeX)));
            } else if (offsetZ) {
                lane = buildMultiplyOp(offsetZ, buildIntVal(min(32, params.localSizeX) * min(32, params.localSizeY)));
            }
            if (lane) {
                lane = buildModOp(
                        buildAddOp(buildVarRefExp("lane_id_" + array, funcScope), lane),
                        buildOpaqueVarRefExp("warpSize", funcScope)
                );
            } else {
                lane = buildVarRefExp("lane_id_" + array, funcScope);
            }
            auto *shfl = buildFunctionCallExp("__shfl", baseType, buildExprListExp(arrRef, lane), funcScope);

            if (offsetX || offsetY || offsetZ) {
                auto *ung = UniqueNameGenerator::getInstance();
                const auto tmp_name = ung->generate("tmp_" + array);

                auto *tmp = buildVariableDeclaration(
                        tmp_name,
                        baseType2,
                        buildAssignInitializer(shfl),
                        funcScope
                );
                auto *border = buildAssignStatement(
                        buildVarRefExp(tmp_name, funcScope),
                        arrRef_copy
                );
                SgIfStmt *border_test = nullptr;
                if (offsetZ) {
                    /*
                     * int tmp = __shfl(local_buffer_input[4], (lane_id_y + j) * 8 + (lane_id_x + i));
                     * if (lane < x * y || lane >= 32 - x*y || lane % x*y < x || lane % x*y == 32 - x
                     *      || lane % x*y % x == x - 1 || lane % x*y % x == 0)
                     * {
                     *     tmp = local_buffer_input[(lidy_input + j) * 3 + (lidx_input + i)];
                     * }
                     * count += tmp;
                     */
                    border_test = buildIfStmt(
                            buildOrOp(
                                    buildOrOp(
                                            buildLessThanOp(
                                                    buildVarRefExp("lane_id_" + array, funcScope),
                                                    buildIntVal(min(32, params.localSizeX * params.localSizeY))
                                            ),
                                            buildGreaterOrEqualOp(
                                                    buildVarRefExp("lane_id_" + array, funcScope),
                                                    buildSubtractOp(
                                                            buildOpaqueVarRefExp("warpSize", funcScope),
                                                            buildIntVal(min(32, params.localSizeX * params.localSizeY))
                                                    )
                                            )
                                    ),
                                    buildOrOp(
                                            buildLessThanOp(
                                                    buildModOp(
                                                            buildVarRefExp("lane_id_" + array, funcScope),
                                                            buildIntVal(min(32, params.localSizeX * params.localSizeY))
                                                    ),
                                                    buildIntVal(min(32, params.localSizeX))
                                            ),
                                            buildOrOp(
                                                    buildGreaterOrEqualOp(
                                                            buildModOp(
                                                                    buildVarRefExp("lane_id_" + array, funcScope),
                                                                    buildIntVal(min(32, params.localSizeX * params.localSizeY))
                                                            ),
                                                            buildSubtractOp(
                                                                    buildOpaqueVarRefExp("warpSize", funcScope),
                                                                    buildIntVal(min(32, params.localSizeX))
                                                            )
                                                    ),
                                                    buildOrOp(
                                                            buildEqualityOp(
                                                                    buildModOp(
                                                                            buildModOp(
                                                                                    buildVarRefExp("lane_id_" + array, funcScope),
                                                                                    buildIntVal(min(32, params.localSizeX * params.localSizeY))
                                                                            ),
                                                                            buildIntVal(min(32, params.localSizeX))
                                                                    ),
                                                                    buildIntVal(min(32, params.localSizeX - 1))
                                                            ),
                                                            buildEqualityOp(
                                                                    buildModOp(
                                                                            buildModOp(
                                                                                    buildVarRefExp("lane_id_" + array, funcScope),
                                                                                    buildIntVal(min(32, params.localSizeX * params.localSizeY))
                                                                            ),
                                                                            buildIntVal(min(32, params.localSizeX))
                                                                    ),
                                                                    buildIntVal(0)
                                                            )
                                                    )
                                            )
                                    )
                            ),
                            border,
                            buildNullStatement());
                } else {
                    /*
                     * int tmp = __shfl(local_buffer_input[4], (lane_id_y + j) * 8 + (lane_id_x + i));
                     * if (lane_id_dx < x || lane_id_dx >= 32 - x || lane_id_dx % x == 0 || lane_id_dx % x == x - 1)
                     * {
                     *     tmp = local_buffer_input[(lidy_input + j) * 3 + (lidx_input + i)];
                     * }
                     * count += tmp;
                     */
                    border_test = buildIfStmt(
                            buildOrOp(
                                    buildOrOp(
                                            buildOrOp(
                                                    buildLessThanOp(
                                                            buildVarRefExp("lane_id_" + array, funcScope),
                                                            buildIntVal(min(32, params.localSizeX))
                                                    ),
                                                    buildGreaterOrEqualOp(
                                                            buildVarRefExp("lane_id_" + array, funcScope),
                                                            buildSubtractOp(
                                                                    buildOpaqueVarRefExp("warpSize", funcScope),
                                                                    buildIntVal(min(32, params.localSizeX))
                                                            )
                                                    )
                                            ),
                                            buildOrOp(
                                                    buildEqualityOp(
                                                            buildDivideOp(
                                                                    buildVarRefExp("lane_id_" + array, funcScope),
                                                                    buildIntVal(min(32, params.localSizeX))
                                                            ),
                                                            buildIntVal(0)
                                                    ),
                                                    buildEqualityOp(
                                                            buildModOp(
                                                                    buildVarRefExp("lane_id_" + array, funcScope),
                                                                    buildIntVal(min(32, params.localSizeX))
                                                            ),
                                                            buildIntVal(0)
                                                    )
                                            )
                                    ),
                                    buildOrOp(
                                            buildEqualityOp(
                                                    buildDivideOp(
                                                            buildVarRefExp("lane_id_" + array, funcScope),
                                                            buildIntVal(min(32, params.localSizeX))
                                                    ),
                                                    buildIntVal(min(32, params.localSizeX) - 1)
                                            ),
                                            buildEqualityOp(
                                                    buildModOp(
                                                            buildVarRefExp("lane_id_" + array, funcScope),
                                                            buildIntVal(min(32, params.localSizeX))
                                                    ),
                                                    buildIntVal(min(32, params.localSizeX) - 1)
                                            )
                                    )
                            ),
                            border,
                            buildNullStatement()
                    );
                }

                SgNode *grandparent = parent->get_parent();
                while (isSgStatement(grandparent) == nullptr) {
                    grandparent = grandparent->get_parent();
                }

                if (auto *grandpa = isSgStatement(grandparent)) {
                    auto *scope = grandpa->get_scope();
                    scope->insert_statement(grandpa, tmp, true);
                    scope->insert_statement(grandpa, border_test, true);
                } else {
                    exit(9);
                }
                parent->replace_expression(arrRef, buildVarRefExp(tmp_name, funcScope));
            } else {
                parent->replace_expression(arrRef, shfl);
            }
        } else {
            std::cerr << "Error in " << __FILE__ << ": parent is not expression!" << std::endl;
            exit(101);
        }
    }

    inline void prependLocalIdComputation(SgBasicBlock *loopBody, const std::string &array) {
        Footprint fp;
        try {
            fp = kernelInfo.getFootprintTable().at(array);
        } catch (const out_of_range &) {
            // If .at(_) throws, we need to see if `localArray` is aliased.
            for (const auto &alias : kernelInfo.aliasArrays) {
                if (array == alias.second.first) {
                    fp = kernelInfo.getFootprintTable().at(alias.first);
                    break;
                }
            }
        }

        const HaloSize hs = fp.computeHaloSize();
        auto *scope = loopBody->get_scope();

        // X
        auto *getLocalIdx = buildAddOp(
                buildVarRefExp("coars_x", scope),
                buildIntVal(hs.left)
        );
        auto *lidxInit = buildAssignInitializer(getLocalIdx);
        auto *lidxDecl = buildVariableDeclaration("lidx_" + array, buildIntType(), lidxInit, scope);

        // Y
        auto *getLocalIdy = buildAddOp(
                buildVarRefExp("coars_y", scope),
                buildIntVal(hs.up)
        );
        auto *lidyInit = buildAssignInitializer(getLocalIdy);
        auto *lidyDecl = buildVariableDeclaration("lidy_" + array, buildIntType(), lidyInit, scope);

        // Z
        if (kernelInfo.isBlockArray(array)) {
            auto *getLocalIdz = buildAddOp(
                    buildVarRefExp("coars_z", scope),
                    buildIntVal(hs.front)
            );
            auto *lidzInit = buildAssignInitializer(getLocalIdz);
            auto *lidzDecl = buildVariableDeclaration("lidz_" + array, buildIntType(), lidzInit, scope);

            loopBody->prepend_statement(lidzDecl);
        }
        loopBody->prepend_statement(lidyDecl);
        loopBody->prepend_statement(lidxDecl);
    }
};

#endif //IMAGECL_REGISTER_CACHING_H
